"""Top-level package for protoattend."""
# fmt: off
__author__ = """Laksh Aithani"""
__email__ = 'lakshaithanii@gmail.com'
__version__ = '0.1.13'
# fmt: on


from protoattend.module import ProtoAttendModule
from protoattend.protoattend import ProtoAttend
